export declare type DiscUnionType<TypeKey extends string> = {
    [M in TypeKey]: string;
};
export declare type Constructor = (...args: any[]) => object;
export declare type Constructors = Record<string, Constructor>;
export declare type WithType<T extends object, K extends string, TypeKey extends string> = Omit<T, TypeKey> & {
    [M in TypeKey]: K;
};
/**
 * Takes an object of type constructors, and returns a discriminated
 * union based on their return types
 * @param T - an object of type constructors (usually the return of discUnion)
 */
export declare type DiscUnionOf<T extends Record<string, (...args: any[]) => object>> = ReturnType<T[keyof T]>;
/**
 * Takes any discriminated union, and narrows the type based
 * on the key
 * @param T - any discriminated union
 * @param K - the name(s) of the types to narrow to
 * @param TypeKey - the name of the discriminant property - defaults to 'type'
 */
export declare type Narrow<T extends DiscUnionType<TypeKey>, K extends T[TypeKey], TypeKey extends string = "type"> = Extract<T, {
    [M in TypeKey]: K;
}>;
/**
 * Takes any discriminated union, and excludes the type of the provided key
 * @param T - any discriminated union
 * @param K - the name(s) of the types to narrow to
 * @param TypeKey - the name of the discriminant property - defaults to 'type'
 */
export declare type Without<T extends DiscUnionType<TypeKey>, K extends string | number | symbol, TypeKey extends string = "type"> = Exclude<T, {
    [M in TypeKey]: K;
}>;
/**
 * Takes any discriminated union, and returns all possible
 * type names as a union of strings
 * @param T - any discriminated union
 * @param TypeKey - the name of the discriminant property - defaults to 'type'
 */
export declare type Keys<T extends DiscUnionType<TypeKey>, TypeKey extends string = "type"> = T[TypeKey];
export declare type ConstructorsWithType<T extends Constructors, TypeKey extends string = "type", Prefix extends string = ''> = {
    [K in keyof T]: {
        <R extends ReturnType<T[K]>>(...args: Parameters<T[K]>): WithType<R, `${Prefix}${Extract<K, string>}`, TypeKey>;
        key: `${Prefix}${Extract<K, string>}`;
    };
};
export declare type Handlers<T extends DiscUnionType<TypeKey>, R, TypeKey extends string = "type"> = {
    [K in T[TypeKey]]: (s: Narrow<T, K, TypeKey>) => R;
};
/**
 * Factory function that creates the disc-union functions. Use this
 * to have a different default discriminant than 'type'
 * @param factoryTypeKey - The default discriminant property of the functions
 */
export declare const factory: <FactoryTypeKey extends string>(factoryTypeKey: FactoryTypeKey) => {
    discUnion: <T extends Constructors, P extends string = "", TypeKey extends string | FactoryTypeKey = FactoryTypeKey>(constructors: T, typeKey?: TypeKey, prefix?: P) => ConstructorsWithType<T, TypeKey, P>;
    match: {
        <T_1 extends DiscUnionType<TypeKey_1>, R extends Handlers<T_1, unknown, TypeKey_1>, TypeKey_1 extends string | FactoryTypeKey = FactoryTypeKey>(value: T_1, handlers: R, typeKey?: TypeKey_1 | undefined): R extends Handlers<T_1, infer RT, TypeKey_1> ? RT : never;
        <T_2 extends DiscUnionType<TypeKey_2>, R_1 extends Partial<Handlers<T_2, unknown, TypeKey_2>>, E, TypeKey_2 extends string | FactoryTypeKey = FactoryTypeKey>(value: T_2, handlers: keyof R_1 extends Keys<T_2, TypeKey_2> ? R_1 : never, otherwise: (value: Exclude<T_2, { [M in TypeKey_2]: keyof R_1; }>) => E, typeKey?: TypeKey_2 | undefined): E | (R_1 extends Partial<Handlers<T_2, infer RT_1, TypeKey_2>> ? RT_1 : never);
    };
    map: <T_3 extends DiscUnionType<TypeKey_3>, K extends T_3[TypeKey_3], R_2, TypeKey_3 extends string | FactoryTypeKey = FactoryTypeKey>(type: K, obj: T_3, mapper: (o: Extract<T_3, { [M_1 in TypeKey_3]: K; }>) => R_2, typeKey?: TypeKey_3) => T_3 | R_2;
    get: {
        <T_4 extends DiscUnionType<TypeKey_4>, K_1 extends T_4[TypeKey_4], TypeKey_4 extends string = FactoryTypeKey>(type: K_1, obj: Extract<T_4, { [M_2 in TypeKey_4]: K_1; }>, typeKey?: TypeKey_4 | undefined): Extract<T_4, { [M_2 in TypeKey_4]: K_1; }>;
        <T_5 extends DiscUnionType<TypeKey_5>, K_2 extends T_5[TypeKey_5], TypeKey_5 extends string | FactoryTypeKey = FactoryTypeKey>(type: K_2, obj: T_5, typeKey?: TypeKey_5 | undefined): Extract<T_5, { [M_3 in TypeKey_5]: K_2; }> | null;
    };
    is: <T_6 extends DiscUnionType<TypeKey_6>, K_3 extends T_6[TypeKey_6], TypeKey_6 extends string | FactoryTypeKey = FactoryTypeKey>(type: K_3, value: T_6, typeKey?: TypeKey_6) => value is Extract<T_6, { [M_4 in TypeKey_6]: K_3; }>;
    validate: <T_7 extends DiscUnionType<TypeKey_7>, K_4 extends T_7[TypeKey_7], TypeKey_7 extends string | FactoryTypeKey = FactoryTypeKey>(type: K_4, obj: T_7, typeKey?: TypeKey_7) => Extract<T_7, { [M_5 in TypeKey_7]: K_4; }>;
    createType: <T_8 extends string, O extends object, TypeKey_8 extends string | FactoryTypeKey = FactoryTypeKey>(type: T_8, obj: O, typeKey?: TypeKey_8) => WithType<O, T_8, TypeKey_8>;
};
/**
 * Accepts an object of constructor functions,
 * and wraps those functions to add a type property
 * to their result
 *
 * @function
 * @param constructors - Object containing constructors
 * @param typeKey - Discriminant property
 */
export declare const discUnion: <T extends Constructors, P extends string = "", TypeKey extends string = "type">(constructors: T, typeKey?: TypeKey, prefix?: P) => ConstructorsWithType<T, TypeKey, P>;
/**
 * Returns the result of mapper if value is of the specified type,
 * otherwise returns the value unchanged
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to map over
 * @param mapper - Function to be called on obj
 * @param typeKey - Discriminant property
 */
export declare const map: <T extends DiscUnionType<TypeKey>, K extends T[TypeKey], R, TypeKey extends string = "type">(type: K, obj: T, mapper: (o: Extract<T, { [M in TypeKey]: K; }>) => R, typeKey?: TypeKey) => T | R;
/**
 * Takes an object to match, and an object of handlers whose
 * keys are the possible types of the discriminated union.
 *
 * Exhaustive unless an "otherwise" handler is passed in as a third argument,
 * in which case the handler object may be partial.
 *
 * @function
 * @param obj - Object to match
 * @param handlers - Object whose keys are possible types of the discriminated union, and whose values are handlers for those types
 * @param otherwise (optional) - Handler if no match is found
 * @param typeKey (optional) - Discriminant property
 * @returns Result of the matched handler, or the result of the "otherwise" handler
 */
export declare const match: {
    <T extends DiscUnionType<TypeKey>, R extends Handlers<T, unknown, TypeKey>, TypeKey extends string = "type">(value: T, handlers: R, typeKey?: TypeKey | undefined): R extends Handlers<T, infer RT, TypeKey> ? RT : never;
    <T_2 extends DiscUnionType<TypeKey_2>, R_2 extends Partial<Handlers<T_2, unknown, TypeKey_2>>, E, TypeKey_2 extends string = "type">(value: T_2, handlers: keyof R_2 extends Keys<T_2, TypeKey_2> ? R_2 : never, otherwise: (value: Exclude<T_2, { [M in TypeKey_2]: keyof R_2; }>) => E, typeKey?: TypeKey_2 | undefined): E | (R_2 extends Partial<Handlers<T_2, infer RT_1, TypeKey_2>> ? RT_1 : never);
};
/**
 * Narrows the type of the value if it matches the specified type,
 * otherwise returns null
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to get
 * @param typeKey - Discriminant property
 */
export declare const get: {
    <T extends DiscUnionType<TypeKey>, K extends T[TypeKey], TypeKey extends string = "type">(type: K, obj: Extract<T, { [M in TypeKey]: K; }>, typeKey?: TypeKey | undefined): Extract<T, { [M in TypeKey]: K; }>;
    <T_2 extends DiscUnionType<TypeKey_2>, K_2 extends T_2[TypeKey_2], TypeKey_2 extends string = "type">(type: K_2, obj: T_2, typeKey?: TypeKey_2 | undefined): Extract<T_2, { [M_1 in TypeKey_2]: K_2; }> | null;
};
/**
 * Determines if the value matches the specified type
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to check
 * @param typeKey - Discriminant property
 */
export declare const is: <T extends DiscUnionType<TypeKey>, K extends T[TypeKey], TypeKey extends string = "type">(type: K, value: T, typeKey?: TypeKey) => value is Extract<T, { [M in TypeKey]: K; }>;
/**
 * Narrows the type of the value if it matches, throws
 * an error otherwise
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to check
 * @param typeKey - Discriminant property
 */
export declare const validate: <T extends DiscUnionType<TypeKey>, K extends T[TypeKey], TypeKey extends string = "type">(type: K, obj: T, typeKey?: TypeKey) => Extract<T, { [M in TypeKey]: K; }>;
/**
 * Convenience function for adding a type discriminant
 * to an object
 *
 * @function
 * @param type - Type to create
 * @param obj - Value to use
 * @param typeKey - Discriminant property
 */
export declare const createType: <T extends string, O extends object, TypeKey extends string = "type">(type: T, obj: O, typeKey?: TypeKey) => WithType<O, T, TypeKey>;
