"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createType = exports.validate = exports.is = exports.maybeGet = exports.match = exports.map = exports.discUnion = exports.factory = void 0;
/**
 * Factory function that creates the disc-union functions. Use this
 * to have a different default discriminant than 'type'
 * @param factoryTypeKey - The default discriminant property of the functions
 */
var factory = function (factoryTypeKey) {
    function match(value, handlers, ot, tk) {
        var _a, _b;
        var typeKey;
        var otherwise;
        if (typeof ot === "function") {
            otherwise = ot;
            typeKey = tk !== null && tk !== void 0 ? tk : factoryTypeKey;
        }
        else {
            typeKey = ot !== null && ot !== void 0 ? ot : factoryTypeKey;
        }
        return ((_b = (_a = handlers[value[typeKey]]) === null || _a === void 0 ? void 0 : _a.call(handlers, value)) !== null && _b !== void 0 ? _b : otherwise(value));
    }
    var is = function (type, value, typeKey) {
        if (typeKey === void 0) { typeKey = factoryTypeKey; }
        return value[typeKey] === type;
    };
    var discUnion = function (constructors, typeKey) {
        if (typeKey === void 0) { typeKey = factoryTypeKey; }
        var constructorsWithType = {};
        var _loop_1 = function (key) {
            constructorsWithType[key] = function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                return createType(key, constructors[key].apply(constructors, args), typeKey);
            };
        };
        for (var key in constructors) {
            _loop_1(key);
        }
        return constructorsWithType;
    };
    var createType = function (type, obj, typeKey) {
        var _a;
        if (typeKey === void 0) { typeKey = factoryTypeKey; }
        return (__assign(__assign({}, obj), (_a = {}, _a[typeKey] = type, _a)));
    };
    var validate = function (type, obj, typeKey) {
        if (typeKey === void 0) { typeKey = factoryTypeKey; }
        if (is(type, obj, typeKey))
            return obj;
        else
            throw new Error("Expected object of type \"" + type + "\", but got \"" + obj[typeKey] + "\"");
    };
    function maybeGet(type, obj, typeKey) {
        if (typeKey === void 0) { typeKey = factoryTypeKey; }
        if (is(type, obj, typeKey))
            return obj;
        else
            return null;
    }
    var map = function (type, obj, mapper, typeKey) {
        if (typeKey === void 0) { typeKey = factoryTypeKey; }
        if (is(type, obj, typeKey))
            return mapper(obj);
        else
            return obj;
    };
    return {
        discUnion: discUnion,
        match: match,
        map: map,
        maybeGet: maybeGet,
        is: is,
        validate: validate,
        createType: createType,
    };
};
exports.factory = factory;
var fns = exports.factory("type");
/**
 * Accepts an object of constructor functions,
 * and wraps those functions to add a type property
 * to their result
 *
 * @function
 * @param constructors - Object containing constructors
 * @param typeKey - Discriminant property
 */
exports.discUnion = fns.discUnion;
/**
 * Returns the result of mapper if value is of the specified type,
 * otherwise returns the value unchanged
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to map over
 * @param mapper - Function to be called on obj
 * @param typeKey - Discriminant property
 */
exports.map = fns.map;
/**
 * Takes an object to match, and an object of handlers whose
 * keys are the possible types of the discriminated union.
 *
 * Exhaustive unless an "otherwise" handler is passed in as a third argument,
 * in which case the handler object may be partial.
 *
 * @function
 * @param obj - Object to match
 * @param handlers - Object whose keys are possible types of the discriminated union, and whose values are handlers for those types
 * @param otherwise (optional) - Handler if no match is found
 * @param typeKey (optional) - Discriminant property
 * @returns Result of the matched handler, or the result of the "otherwise" handler
 */
exports.match = fns.match;
/**
 * Narrows the type of the value if it matches the specified type,
 * otherwise returns null
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to get
 * @param typeKey - Discriminant property
 */
exports.maybeGet = fns.maybeGet;
/**
 * Determines if the value matches the specified type
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to check
 * @param typeKey - Discriminant property
 */
exports.is = fns.is;
/**
 * Narrows the type of the value if it matches, throws
 * an error otherwise
 *
 * @function
 * @param type - Type to match
 * @param obj - Value to check
 * @param typeKey - Discriminant property
 */
exports.validate = fns.validate;
/**
 * Convenience function for adding a type discriminant
 * to an object
 *
 * @function
 * @param type - Type to create
 * @param obj - Value to use
 * @param typeKey - Discriminant property
 */
exports.createType = fns.createType;
